// FUSE file system for vlbi_streamer style recordings
#define FUSE_USE_VERSION 26
#include <fuse.h>

// Project includes
#include <auto_array.h>
#include <evlbidebug.h>
#include <regular_expression.h>
#include <dosyscall.h>
#include <directory_helper_templates.h>

// Standardized C++ headers
#include <iostream>
#include <map>
#include <set>
#include <list>
#include <queue>
#include <string>
#include <exception>
#include <stdexcept>
#include <iterator>
#include <limits>
#include <algorithm>
#include <cstddef>
#include <cerrno>
#include <cstring>
#include <csignal>
#include <cstdlib>
#include <climits>

// Old-style *NIX headers
#include <glob.h>
#include <fnmatch.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/resource.h>


using namespace std;

DECLARE_EZEXCEPT(vbs_except)
DEFINE_EZEXCEPT(vbs_except)

typedef std::vector<std::string>        patternlist_type;
typedef std::vector<Regular_Expression> matchlist_type;
string escape(string const& s);

// Get a handle for invalid file descriptor that isn't -1.
const int    invalidFileDescriptor = std::numeric_limits<int>::min();
const size_t pgSize( ::getpagesize() );

// prototype now, impl follows below
int vbs_pthread_create(pthread_t*, void*(*)(void*), void *arg);
struct mk6_file_header;


/////////////////////////////////////////////////////////////////////////////////////////////
//
//
// This is what it's all about: mapping recording name to a Mk6/VBS
// scattered recording!
//
//
//////////////////////////////////////////////////////////////////////////////////////////////
typedef set<string>                      paths_type;
typedef list<pthread_t>                  threadlist_type;



///////////////////////////////////////////////////////////////
//
//   Update metadata from src => dst
//   in a specific manner
//
///////////////////////////////////////////////////////////////
static const mode_t permissionBits = (S_IRWXU | S_IRWXG | S_IRWXO | S_ISUID | S_ISGID | S_ISVTX);

ostream& operator<<(ostream& os, struct stat const& st) {
    return os << "[st_mode=" << st.st_mode << " uid=" << st.st_uid << " gid=" << st.st_gid << 
                 " sz=" << st.st_size << " st_mtime=" << st.st_mtime << "]";
}

void update(struct stat& dst, struct stat const& src) {
    // dst's st_mtime should be maximum of src/dst's st_mtime
    dst.st_mtime = std::max(dst.st_mtime, src.st_mtime);
    // If mode already filled in don't do that again
    if( dst.st_mode )
        return;
    // Set mode to file + current permissions, owner and group
    dst.st_mode = S_IFREG | (src.st_mode & permissionBits);
    dst.st_uid  = src.st_uid;
    dst.st_gid  = src.st_gid;
}

///////////////////////////////////////////////////////////////
//
// Simple utility: scoped mutex lock
// Always comes in handy
//
///////////////////////////////////////////////////////////////
struct scoped_lock_type {
    scoped_lock_type(pthread_mutex_t* mtx):
        __m_slt_mutex_ptr( mtx )
    { ::pthread_mutex_lock(__m_slt_mutex_ptr); }

    ~scoped_lock_type() {
        ::pthread_mutex_unlock(__m_slt_mutex_ptr);
    }

    pthread_mutex_t* __m_slt_mutex_ptr;
};

struct scoped_readlock_type {
    scoped_readlock_type(pthread_rwlock_t* rw):
        __m_rwlock_ptr( rw )
    { ::pthread_rwlock_rdlock( __m_rwlock_ptr ); }

    ~scoped_readlock_type() {
        ::pthread_rwlock_unlock( __m_rwlock_ptr );
    }
    pthread_rwlock_t*   __m_rwlock_ptr;
};
struct scoped_writelock_type {
    scoped_writelock_type(pthread_rwlock_t* rw):
        __m_rwlock_ptr( rw )
    { ::pthread_rwlock_wrlock( __m_rwlock_ptr ); }

    ~scoped_writelock_type() {
        ::pthread_rwlock_unlock( __m_rwlock_ptr );
    }
    pthread_rwlock_t*   __m_rwlock_ptr;
};

///////////////////////////////////////////////////////////////////////////////
//    Below code copied from Jan Wagner's mark6sg FUSE file system
//    which is copied from Roger Cappallo's d-plane software ...
//
//    [HV: Why are block numbers & sizes ints and not unsigned ints?
//    We will have to follow the convention or else we may construct invalid
//    files. But why????
//    Using unsigned:
//                - gives you another factor of two in block numbers ....
//                - how can block numbers go *negative* in the first place?!?!!
//                - likewise for packet sizes: how could they ever be
//                  _negative_???
//
//           Also made the ints of actual 32-bit size; you can't really
//           leave it up to the compiler to decide how long your int really
//           is if you want portable, 64-bit clean code]


// Internal structs
//
// There is apparently no documentation on the Mark6 non-RAID recording format(?).
// From Mark6 source code one can see that the several files associated with a
// single scan all look like this:
//
//  [file header]
//  [block a header] [block a data (~10MB)]
//  [block b header] [block b data (~10MB)]
//  [block c header] [block c data (~10MB)]
//  ...
//
// Block numbers are increasing. They will have gaps (e.g, a=0, b=16, c=35, ...).
// The 'missing' blocks (1,2,3..,15,17,18,..,34,...) of one file are found in one of
// the other files. Blocks can have different data sizes.
//
// Because 10GbE recording in Mark6 software is not done Round-Robin across files,
// the order of blocks between files is somewhat random.
//
// The below structures are adopted from the source code of the
// Mark6 program 'd-plane' version 1.12:


#define MARK6_SG_SYNC_WORD          0xfeed6666


struct mk6_file_header {             // file header - one per file
    // HV: Stick the enum in the mk6_file_header struct such that
    //     the global namespace doesn't get clobbered.
    enum packet_formats {
        VDIF,
        MK5B,
        UNKNOWN_FORMAT
    };
    uint32_t  sync_word;              // MARK6_SG_SYNC_WORD
    int32_t   version;                // defines format of file
    int32_t   block_size;             // length of blocks including header (bytes)
    int32_t   packet_format;          // format of data packets, enumerated below
    int32_t   packet_size;            // length of packets (bytes)
};

struct mk6_wb_header_v2 {           // write block header - version 2
    int32_t blocknum;               // block number, starting at 0
    int32_t wb_size;                // same as block_size, or occasionally shorter
};



///////////////////////////////////////////////////
//
//   A filedescriptor cache
//
// Mark6 scatters chunks over a number of files
// so many chunks share the same file descriptor.
// After opening a recording then the path opener
// is called for each path of the recording.
//
// The Mk6 opener pre-opens all files for reading
// and puts them in the cache. The vbs opener doesn't.
//
// Then, if a chunk needs to be read from disk,
// the reader checks the cache for the fd. If found,
// it uses that file descriptor. Otherwise it temporarily
// opens the chunks file path, reads data and closes
// it again.
//
///////////////////////////////////////////////////

typedef map<string, int>    fdcache_type;
typedef int(*path_opener_fn)(string const& p);

fdcache_type                fdcache;
pthread_rwlock_t            fdcache_lock = PTHREAD_RWLOCK_INITIALIZER;

// Pointer to a path opener function.
typedef int (*openpath_fn)(string const& p, time_t& tmHint, off_t& szHint);
typedef int (*closepath_fn)(string const& p);

// Mk6 path opener puts opened file in the fdcache
int mk6_path_opener(string const& p, time_t&, off_t&) {
    int   fd = ::open(p.c_str(), O_RDONLY );
    if( fd<0 ) {
        DEBUG(-1, "mk6_path_opener: failed to open '" << p << "' - " << ::strerror(errno) << endl);
        return -1;
    }
    // And put it in the cache
    scoped_writelock_type  swl( &fdcache_lock );
    return fdcache.insert(make_pair(p, fd)).second==true ? 0 : -1;
}

int mk6_path_closer(string const& p) {
    scoped_writelock_type swl( &fdcache_lock );

    fdcache_type::iterator fdptr = fdcache.find(p);
    if( fdptr==fdcache.end() ) {
        DEBUG(-1, "mk6_path_closer: path '" << p << "' not in cache?!" << endl);
        return -1;
    }
    ::close( fdptr->second );
    fdcache.erase( fdptr );
    return 0;
}

// vbs path opener doesn't do anything
int vbs_path_opener(string const&, time_t&, off_t& ) {
    return 0;
}
int vbs_path_closer(string const& ) {
    return 0;
}

///////////////////////////////////////////////////
//
//   A mountpoint monitor's private parts
//
///////////////////////////////////////////////////
struct filechunk_type;

typedef set<int>                observers_type;
typedef queue<filechunk_type*>  workqueue_type;

struct mphandler_data {
    bool             stop;
    bool             indexed;
    pthread_t        threadId;
    const string     mountPoint;
    paths_type       matchList;
    workqueue_type   workqueue;
    pthread_cond_t*  condition;
    pthread_mutex_t* mutex;

    mphandler_data(string const& mp, paths_type const& indexPatterns):
        stop( false ), indexed( false ), mountPoint( mp ), matchList( indexPatterns ),
        condition( new pthread_cond_t ), mutex( new pthread_mutex_t )
    {
        // Initialize the mutex and condition variable
        ::pthread_cond_init(condition, 0);
        ::pthread_mutex_init(mutex, 0);
    }

    // Implement operator()(string const&) => we can act as a filter!
    bool operator()(string const& nm) const {
        // the default is wether we need to filter or not
        //   matchList empty?      No filtering => let everything pass
        //   matchList not empty?  Start with false value and if something
        //                          matches let the thing pass
        bool   rv( matchList.empty() );
        for( paths_type::const_iterator curRX=matchList.begin(); curRX!=matchList.end(); curRX++) {
            if( (rv=(::fnmatch(curRX->c_str(), nm.c_str(), FNM_PATHNAME)==0))==true )
                break;
        }
        return rv;
    }
};

// Keep a global mapping of recording => filechunks
// note that we let the filechunks be an automatically sorted container
mphandler_data& get_mpdata( string const& mp );

struct filechunk_type {
    // Note: no default c'tor


    // VBS chunk:
    //     create from mountpoint, <recname/recname>.<sequencenr>, <sequencenr>, <size>
    //
    // Mk6 chunk:
    //     create from mountpoint, <recname>, <sequencenr>, <size>, <fileOffset>


    // construct from mountpoint AND full path name
    // the mountpoint will be used as lookup key for the thread to ask for the chunk's data
    filechunk_type(string const& mp, string const& fnm, unsigned int seqno, off_t sz, off_t fpos = 0) :
        mountPoint( mp ), pathToChunk( fnm ), chunkSize( sz ), chunkPos( fpos ), chunkOffset( 0 ), chunkNumber( seqno ),
        __m_saved_errno( ESRCH ), __m_thread_active( false ), __m_priority( 0 ),
        __m_bytes( 0 ), __m_cond( new pthread_cond_t ), __m_mutex( new pthread_mutex_t ), __m_rwlock( new pthread_rwlock_t )
    {
        // Initialize the mutex and condition variable
        ::pthread_cond_init(__m_cond, 0);
        ::pthread_mutex_init(__m_mutex, 0);
        ::pthread_rwlock_init(__m_rwlock, 0);
    }


    // Note: declare chunkOffset as mutable such that we can later, after
    // the chunks have been sorted and put into a set, update the
    // chunkOffset value to what it should be. [Elements in a set are const
    // in order to guarantee that their sorting order is not compromised
    // as you alter the element - in this case WE know that the sorting only
    // depends on the value of 'chunkNumber' so we can safely edit
    // chunkOffset w/o worrying about compromising the set]
    string          mountPoint;
    string          pathToChunk;
    off_t           chunkSize;
    off_t           chunkPos;    // chunk's startposition in file
    mutable off_t   chunkOffset; // offset within recording
    unsigned int    chunkNumber;
    //struct stat     chunkStat;

    // Given a file pointer which must lay at or beyond the start
    // of this chunk. If pointing outside the far end, return 0
    // [this is typically true if someone seeks outside the file
    // size's range or if we've reached EOF but wait for further
    // incoming data].
    size_t remainingBytes(off_t fp) const {
        const off_t limit = chunkOffset + chunkSize;
        return (size_t)((fp>=chunkOffset && fp<=limit) ? (limit - fp) : 0);
    }

    // (pre-)reading API of the chunk.
    // It is possible to initiate reading a chunk's data and then later wait
    // for it to become available

    // This initiates a thread, reading the chunk's data, if it's not
    // already in memory. This method may safely be called multiple times,
    // although NOT from DIFFERENT THREADS at the same time on the same
    // chunk - make sure you hold the chunk lock.
    void initiateRead( int observer, unsigned int priority ) const {
        // Can used scoped lock
        scoped_lock_type    slt( __m_mutex );

        DEBUG(2, "iniateRead[" << mountPoint << "/" << pathToChunk << ", prio=" << priority << "]" << endl);

        // No observers => data isn't read yet, so must issue a request
        if( __m_observers.empty() || priority>__m_priority ) {
            scoped_writelock_type   swl( __m_rwlock );

            // if observers not empty, requested priority > current prio,
            // i.e. we must replace current mmapped bytes with block read
            // from disk
            if( !__m_observers.empty() ) {
                // Note: mmap() can only map on page-size boundaries
                //       so the /actual/ pointer returned from mmap
                //       can be computed by shifting it back by the
                //       amount the /actual/ "chunk" started within
                //       the page
                ::munmap( __m_bytes - (chunkPos % pgSize), chunkSize );
                __m_bytes = 0;
            }
            // Can already switch to new priority
            __m_priority = priority;

            // Depending on prio, do the right thing
            if( __m_priority==0 ) {
                // Handle the mmapping inside this call w/o giving up locks
                filechunk_type::mmap_chunk(const_cast<filechunk_type*>(this));
            } else {
                // Append ourselves to the appropriate workqueue - the queue
                // for the mountpoint we're located on!
                mphandler_data&   mpdata( get_mpdata(mountPoint) );
                scoped_lock_type  wqlock( mpdata.mutex );

                // Adding to the workqueue is not enough (guess how I found this out!)
                // we must also signal the condition - how else is the worker thread
                // to know that something happened?!
                if( !mpdata.stop ) {
                    mpdata.workqueue.push( const_cast<filechunk_type*>(this) );
                    ::pthread_cond_broadcast(mpdata.condition);
                }
                // Also indicate that we're being read
                __m_thread_active = true;
            }
        }

        // Insert observer into set
        __m_observers.insert( observer );
    }

    // At some time after init you call this to wait indefinitely for the data to be
    // in memory. If it's already there it returns quickly, obviously.
    //
    // If it returns NULL, it sets errno as follows:
    //   * ESRCH  'No such process' if no call to initateRead() was made
    //     first
    //   * ENOMEM if the memory allocation for this file chunk failed
    //   * any of the errno's returned by open(2) or read(2), if a failure
    //     occurred during any of these operations, trying to read the 
    //     chunk's data into memory
    //
    unsigned char* readLockForData( void ) const {
        scoped_lock_type    slt( __m_mutex );

        // Wait whilst no data & thread active.
        // As soon as either is false, we can do stuff
        while( __m_bytes==0 && __m_thread_active )
            ::pthread_cond_wait(__m_cond, __m_mutex);

        if( __m_bytes==0 )
            errno = __m_saved_errno;
        else
            // grab readlock on the data
            ::pthread_rwlock_rdlock( __m_rwlock );
        return __m_bytes;
    }
    void releaseReadLock( void ) const {
        ::pthread_rwlock_unlock( __m_rwlock );
    }

    // When the caller has no use for this chunk's data anymore, please to 
    // release it!
    void releaseData( int observer ) const {
        scoped_lock_type         slt( __m_mutex );
        const size_t             nobs   = __m_observers.size();
        observers_type::iterator obsptr = __m_observers.find( observer );

        // Check if the observer was observing this block
        if( obsptr==__m_observers.end() )
            return;

        // Remove it from the set. If the set of observers is empty, then we can safely 
        // deallocate the block.
        __m_observers.erase( obsptr );

        DEBUG(2, "filechunk[" << pathToChunk << "]::releaseData - data:" << (__m_bytes!=0)
                 << " nObservers:" << nobs << " -> " << __m_observers.size() << endl);

        // still observers left - we're done
        if( __m_observers.size() )
            return;

        // No more referrers?
        // Then we must at least wait for a possible thread to finish
        // [E.g. file closed before read-ahead chunk was actually needed]
        while( __m_thread_active )
            ::pthread_cond_wait(__m_cond, __m_mutex);

        // Good, now that the thread is gone, we can delete the data
        // Do that depending on how it was loaded
        if( __m_bytes ) {
            if( __m_priority==0 )
                // Note: mmap() can only map on page-size boundaries
                //       so the /actual/ pointer returned from mmap
                //       can be computed by shifting it back by the
                //       amount the /actual/ "chunk" started within
                //       the page
                ::munmap( __m_bytes - (chunkPos % pgSize), chunkSize );
            else
                ::free( __m_bytes );
        }
        __m_bytes       = 0;
        __m_saved_errno = ESRCH;
    }

    private:
        // the mountpoint monitor thread function has
        // access to our internals
        friend void* mountpoint_monitor(void*);

        // no default c'tor!
        filechunk_type();

        // the data + mutex &cet protecting the object

        mutable int               __m_saved_errno;
        mutable bool              __m_thread_active;
        mutable unsigned int      __m_priority; // 0 = mmap(), >0 = read_chunk()/malloc thus ::free()
        mutable observers_type    __m_observers; 
        mutable unsigned char*    __m_bytes;
        mutable pthread_cond_t*   __m_cond;
        mutable pthread_mutex_t*  __m_mutex;
        mutable pthread_rwlock_t* __m_rwlock;

        // Minor wrapper arounce read_chunk that detaches.
        // Best be executed inna thread
        static void* mmap_chunk(filechunk_type* fcptr) {
            int             fd = invalidFileDescriptor, e;
            bool            closeFd( true );
            const off_t     pgAligned( (fcptr->chunkPos/pgSize)*pgSize );
            const size_t    pgOffset( fcptr->chunkPos%pgSize );
            unsigned char*  bytes = 0;

            DEBUG(3, "mmap_chunk[" << ::pthread_self() << "] mapping " << fcptr->pathToChunk << ":" << fcptr->chunkPos
                     << " + " << fcptr->chunkSize << "bytes [pgAligned: " << pgAligned << " + " << fcptr->chunkSize+pgOffset << "bytes]" << endl);

            // Check if file dscriptor already open
            {
                scoped_readlock_type   srl( &fdcache_lock );
                fdcache_type::iterator fdptr = fdcache.find( fcptr->pathToChunk );

                if( fdptr!=fdcache.end() ) {
                    fd      = fdptr->second;
                    closeFd = false;
                }
            }

            // If fd still invalid, then we have to open the file ourselves :-(
            if( fd==invalidFileDescriptor ) {
                // If we cannot open the file ...
                if( (fd=::open(fcptr->pathToChunk.c_str(), O_RDONLY))<0 ) {
                    e = errno;
                    DEBUG(-1, "mmap_chunk[" << fcptr->pathToChunk << "] - failed to open: " << ::strerror(errno) << endl);
                    fcptr->__m_saved_errno   = e;
                    fcptr->__m_thread_active = false;
                    fcptr->__m_bytes         = 0;
                    return (void*)0;
                }
                closeFd = true;
            }

            // HV: Hmmm. We can only mmap page-aligned offsets from within the file
            //     So the amount of bytes to mmap() has to grow by the
            //     amount of bytes we have shifted the actual pgAligned
            //     address downward
            bytes = (unsigned char*)::mmap((void*)0, (size_t)fcptr->chunkSize + pgOffset, PROT_READ, MAP_PRIVATE, fd, pgAligned);
            // Save errno for possible later use
            e     = errno;

            if( bytes==MAP_FAILED ) {
                DEBUG(-1, "mmap_chunk[" << ::pthread_self() << "] mmap failed on " <<
                          fcptr->pathToChunk << ":" << fcptr->chunkPos << " + " << fcptr->chunkSize <<
                          " [pgAligned: " << pgAligned << " + " << (fcptr->chunkSize+pgOffset) << "]" <<
                          " - " << ::strerror(e) << endl);
            }
            fcptr->__m_saved_errno   = (bytes==MAP_FAILED) ? e : 0;
            fcptr->__m_thread_active = false;
            fcptr->__m_bytes         = (bytes==MAP_FAILED) ? 0 : (bytes + pgOffset);

            // If we need to close the file descriptor, do it now
            (void)(closeFd && ::close( fd ));
            return (void*)0;
        }

        // chunk-reading function. preferrably to be executed
        // inna different thread of execution
        static void* read_chunk(void* args) {
            int                    fd = invalidFileDescriptor, e = 0;
            bool                   closeFd( true );
            unsigned char*         bytes = 0;
            filechunk_type*        fcptr = (filechunk_type*)args;

            DEBUG(2, "read_chunk[" << ::pthread_self() << "] reading " << fcptr->pathToChunk << ", " <<
                     fcptr->chunkSize << " bytes" << endl);

            // Check if file dscriptor already open
            {
                scoped_readlock_type   srl( &fdcache_lock );
                fdcache_type::iterator fdptr = fdcache.find( fcptr->pathToChunk );

                if( fdptr!=fdcache.end() ) {
                    fd      = fdptr->second;
                    closeFd = false;
                }
            }

            // If fd still invalid, then we have to open the file ourselves :-(
            if( fd==invalidFileDescriptor ) {
                // If we cannot open the file ...
                if( (fd=::open(fcptr->pathToChunk.c_str(), O_RDONLY))<0 ) {
                    e = errno;
                    DEBUG(-1, "read_chunk[" << fcptr->pathToChunk << "] - failed to open: " << ::strerror(errno) << endl);
                    ::pthread_mutex_lock(fcptr->__m_mutex);
                    fcptr->__m_saved_errno   = e;
                    fcptr->__m_thread_active = false;
                    fcptr->__m_bytes         = 0;
                    ::pthread_cond_broadcast(fcptr->__m_cond);
                    ::pthread_mutex_unlock(fcptr->__m_mutex);
                    return (void*)0;
                }
                closeFd = true;
            }

            // use ::malloc() because it doesn't throw - we cannot throw
            // exceptions across threads and this executes inna different
            // thread
            if( (bytes=(unsigned char*)::malloc(fcptr->chunkSize))==0 ) {
                (void)(closeFd && ::close( fd ));
                DEBUG(-1, "read_chunk[" << fcptr->pathToChunk << "] - insufficient memory for " << fcptr->chunkSize << " bytes" << endl);
                ::pthread_mutex_lock(fcptr->__m_mutex);
                fcptr->__m_saved_errno   = ENOMEM;
                fcptr->__m_thread_active = false;
                fcptr->__m_bytes         = 0;
                ::pthread_cond_broadcast(fcptr->__m_cond);
                ::pthread_mutex_unlock(fcptr->__m_mutex);
                return (void*)0;
            }

            // seek to the chunk's position in the file
            ::lseek(fd, fcptr->chunkPos, SEEK_SET);

            // Nothing left than to read the data
            size_t  n = (size_t)fcptr->chunkSize;
            while( n ) {
                ssize_t nr = ::read(fd, bytes+(fcptr->chunkSize-n), n);
                if( nr<=0 ) {
                    if( nr<0 ) {
                        e = errno;
                        DEBUG(-1, "read_chunk[" << fcptr->pathToChunk << "] - failed to read() " << ::strerror(errno) << endl);
                    }
                    break;
                }
                n -= (size_t)nr;
            }
            // If we need to close the file descriptor, do it now
            (void)(closeFd && ::close( fd ));

            // Ok, indicate we're done!
            ::pthread_mutex_lock(fcptr->__m_mutex);
            fcptr->__m_saved_errno   = e;
            fcptr->__m_thread_active = false;
            fcptr->__m_bytes         = bytes;
            ::pthread_cond_broadcast(fcptr->__m_cond);
            ::pthread_mutex_unlock(fcptr->__m_mutex);

            DEBUG(2, "read_chunk[" << ::pthread_self() << "] DONE reading " << fcptr->pathToChunk << endl);
            return (void*)0;
        }
};
typedef set<filechunk_type>              filechunks_type;

// Comparison operator for filechunk_type - sort by chunkNumber exclusively!
bool operator<(filechunk_type const& l, filechunk_type const& r) {
    return l.chunkNumber < r.chunkNumber;
}

// And we keep metadata per recording separately
struct metadata_type {
    off_t       recordingSize;
    struct stat st;         // used for permissions and st_mtime

    metadata_type():
        recordingSize( 0 )
    { ::memset(&st, 0x0, sizeof(struct stat)); }
};


////////////////////////////////////////////////////////////////
//
//  Prototypes so we can use the calls; implementation is at the
//  bottom of this module
//
////////////////////////////////////////////////////////////////

// These look for VBS recordings
void scanVBSRecordingDirectory(string const& recname, string const& dir, filechunks_type& fcs, struct stat& ts);
void scanMk6RecordingFile(string const& recname, string const& file, filechunks_type& fcs, struct stat& ts);

typedef void (*scanfunction_ptr)(string const& recname, string const& path, filechunks_type& fc, struct stat& tsptr);

// When can we switch to c++11? This sucks, having to do pthread stuff!
struct scanfunction_args {
    string           path;
    string           recordingName;
    struct stat*     stPtr;
    pthread_mutex_t* mtx;
    scanfunction_ptr sfPtr;
    filechunks_type* fcPtr;

    scanfunction_args(string const& rec, string const& p, scanfunction_ptr sf,
                      filechunks_type* fcptr, struct stat* stptr, pthread_mutex_t* m):
        path( p ), recordingName( rec ), stPtr( stptr ), mtx( m ), sfPtr( sf ), fcPtr( fcptr )
    {}

    private:
        scanfunction_args();
        scanfunction_args(const scanfunction_args&);
        const scanfunction_args& operator=(const scanfunction_args&);
};

void* doScan(void* ptr) {
    scanfunction_args*   sfargs( (scanfunction_args*)ptr );
    try {
        // Accumulate file chunks in local variable
        struct stat     tmpST;
        filechunks_type tmpFC;

        ::memset(&tmpST, 0x0, sizeof(struct stat));
        sfargs->sfPtr(sfargs->recordingName, sfargs->path, tmpFC, tmpST);

        // Now lock the result and transfer our results into that
        scoped_lock_type                sml( sfargs->mtx );

        std::copy(tmpFC.begin(), tmpFC.end(), std::insert_iterator<filechunks_type>(*sfargs->fcPtr, sfargs->fcPtr->begin()));
        update(*sfargs->stPtr, tmpST);
    }
    catch( int e ) {
        DEBUG(-1, "doScan(" << sfargs->path << ") - errno=" << e << " [" << ::strerror(e) << endl);
    }
    catch( std::exception& e ) {
        DEBUG(-1, "doScan(" << sfargs->path << ") - exception " << e.what() << endl);
    }
    catch( ... ) {
        DEBUG(-1, "doScan(" << sfargs->path << ") - unknown exception" << endl);
    }
    delete sfargs;
    return (void*)0;
}



// Describes a recording
struct recording_type {
    const string            recordingName; // Const because this will be the primary key!
    mutable paths_type      paths;         // Mk6 => files, VBS => directories
    mutable metadata_type   metadata;
    mutable observers_type  observers;
    pthread_mutex_t         mutex;
    mutable filechunks_type fileChunks;
    scanfunction_ptr        scanFunction;
    openpath_fn             pathOpener;
    closepath_fn            pathCloser;


    recording_type(recording_type const& other):
        recordingName( other.recordingName ), scanFunction( other.scanFunction ), 
        pathOpener( other.pathOpener ), pathCloser( other.pathCloser )
    {
        ::pthread_mutex_init(&mutex, 0);
    }

    recording_type(string const& recname, scanfunction_ptr sf, openpath_fn opener, closepath_fn closer):
        recordingName( recname ), scanFunction( sf ), pathOpener( opener ), pathCloser( closer )
    { 
        ::pthread_mutex_init(&mutex, 0);
    }

    void add_path(string const& p) const {
        scoped_lock_type  sml( const_cast<pthread_mutex_t*>(&mutex) );

        if( paths.insert(p).second==false )
            throw runtime_error(string("Duplicate insert for path '")+p+"'");
        // Can have quick look at path - get modification time estimate?
        // in case it's a file we then might also get an estimated file size
        // and user/group id?
    }

    int open(int observer) const {
        scoped_lock_type  sml( const_cast<pthread_mutex_t*>(&mutex) );

        // Add observer (do not allow duplicates)
        if( observers.insert(observer).second==false ) {
            DEBUG(-1, "recording/open: duplicate observer " << observer << endl);
            return -1;
        }
        // Must be indexed
        int   e;

        if( (e=doIndex_unlocked())!=0 )
            return e;
        // And open necessary files
        time_t t;
        off_t  o;
        for(paths_type::iterator pptr=paths.begin(); pptr!=paths.end(); pptr++)
            if( (e=pathOpener(*pptr, t, o))!=0 )
                break;
        return e;
    }

    int index( void ) const {
        scoped_lock_type  sml( const_cast<pthread_mutex_t*>(&mutex) );
        return doIndex_unlocked();
    }

    int close(int observer) const {
        scoped_lock_type  sml( const_cast<pthread_mutex_t*>(&mutex) );
        
        observers_type::iterator  optr = observers.find(observer);
        if( optr==observers.end() ) {
            DEBUG(-1, "recording/close: observer " << observer << " not in list!" << endl);
            return -1;
        }
        observers.erase( optr );

        // No one left? Then we need to close all paths
        if( !observers.empty() )
            return 0;
        // Note that we do not break on error - we must attempt to close
        // all paths
        int e = 0;
        for(paths_type::iterator pptr=paths.begin(); pptr!=paths.end(); pptr++) {
            int tmpe = pathCloser( *pptr );
            // remember first failure
            if( tmpe && !e )
                e = tmpe;
        }
        return e;
    }

    private:
        int doIndex_unlocked( void ) const {
            if( fileChunks.size() )
                return 0;

            // Oh bugger
            // Issue a multithreaded parallel scan
            int              eno = 0;
            pthread_mutex_t  mtx = PTHREAD_MUTEX_INITIALIZER;
            threadlist_type  threads;

            // Make sure metadata is reset to empty
            metadata = metadata_type();

            for(paths_type::iterator pptr=paths.begin(); pptr!=paths.end(); pptr++) {
                pthread_t          threadId;
                scanfunction_args* sfaptr(new scanfunction_args(recordingName, *pptr, scanFunction,&fileChunks,
                                                                &metadata.st, &mtx));
                if( (eno=vbs_pthread_create(&threadId, doScan, sfaptr))!=0 )
                    break;
                threads.push_back( threadId );
            }
            // Join all threads
            for(threadlist_type::iterator tidptr=threads.begin(); tidptr!=threads.end(); tidptr++)
                ::pthread_join( *tidptr, 0 );

            // Go through all file chunks, updating their position and also
            // keep track of the recording size
            for(filechunks_type::iterator p=fileChunks.begin(); p!=fileChunks.end(); p++) {
                p->chunkOffset = metadata.recordingSize;
                metadata.recordingSize += p->chunkSize;
            }
            return eno;
        }

        // None of these should be allowed
        recording_type();
        recording_type const& operator=(recording_type const&);
};

// Define strict weak ordering for recording_type: 
//   compare by name only
bool operator<(recording_type const& l, recording_type const& r) {
    return l.recordingName < r.recordingName;
}

typedef set<recording_type>     recordings_type;


// Open file information
struct openfile_type {
    // remember last pointer to filechunk
    off_t                       previousEnd;
    uint64_t                    nConsecutiveReads;
    pthread_rwlock_t            fileRWLock;
    recordings_type::iterator   recordingPtr;
    filechunks_type::iterator   curChunk;

    openfile_type(openfile_type const& other):
        nConsecutiveReads( 0 ), recordingPtr( other.recordingPtr ), curChunk( other.curChunk )
    {
        int r;
        // default rwlock is good enough for us!
        if( (r=::pthread_rwlock_init(&fileRWLock, 0))!=0 ) {
            DEBUG(-1, "openfile_type(openfile_type const&)/pthread_rwlock_init fails - " << ::strerror(r) << endl);
        }
        // Put previous end past end-of-file such that it should, normally,
        // not compare equal to the first read() action at any file offset
        previousEnd = recordingPtr->metadata.recordingSize + 1;
    }

    openfile_type(recordings_type::iterator recptr):
        nConsecutiveReads( 0 ), recordingPtr( recptr ), curChunk( recordingPtr->fileChunks.begin() )
    {
        int r;
        // default rwlock is good enough for us!
        if( (r=::pthread_rwlock_init(&fileRWLock, 0))!=0 ) {
            DEBUG(-1, "openfile_type()/pthread_rwlock_init fails - " << ::strerror(r) << endl);
        }
        // Put previous end past end-of-file such that it should, normally,
        // not compare equal to the first read() action at any file offset
        previousEnd = recordingPtr->metadata.recordingSize + 1;
    }

    private:
        openfile_type();
        openfile_type const& operator=(openfile_type const&);
};



///////////////////////////////////////////////////
//
//   typedefs
//
///////////////////////////////////////////////////

typedef map<string, mphandler_data>     mpmonitormap_type;
typedef map<int, openfile_type>         openfilecache_type;



///////////////////////////////////////////////
//   Keep state of the FUSE file system in
//   a struct of this type. It's a global
//   variable.
///////////////////////////////////////////////
static const string flexbuffPattern( "/mnt/disk*" );
static const string mk6Pattern( "/mnt/disks/*/*/data" );

struct vbs_state_type {
    int                  readAhead;
    bool                 fullIndex;
    paths_type           mountPoints;
    paths_type           indexPatterns;
    recordings_type      recordings;
    pthread_mutex_t      chunkMutex;
    mpmonitormap_type    mpmonitorMap;
    openfilecache_type   openfileCache;

    vbs_state_type():
        readAhead( 0 ), fullIndex( true )
    {
        ::pthread_mutex_init(&chunkMutex, 0);
    }

    ~vbs_state_type() {
        DEBUG(2, "~vbs_state_type: stopping all monitor threads" << endl);
        // Stop all mountpoint monitors
        for( mpmonitormap_type::iterator monptr=mpmonitorMap.begin(); monptr!=mpmonitorMap.end(); monptr++) {
            mphandler_data&      mpdata = monptr->second;

            DEBUG(2, "~vbs_state_type: stopping " << monptr->first << endl);

            // Signal thread that it's time to go
            ::pthread_mutex_lock(mpdata.mutex);
            mpdata.stop = true;
            ::pthread_cond_signal(mpdata.condition);
            ::pthread_mutex_unlock(mpdata.mutex);
        }


        for( mpmonitormap_type::iterator monptr=mpmonitorMap.begin(); monptr!=mpmonitorMap.end(); monptr++) {
            mphandler_data&      mpdata = monptr->second;
            // Now join it
            int   j;
            void* dummy;

            DEBUG(2, "~vbs_state_type: joining " << monptr->first << endl);
            j = ::pthread_join(mpdata.threadId, &dummy);
            if( j!=0 ) {
                DEBUG(-1, "~vbs_state_type: failed to join [" << monptr->first << "]: " << ::strerror(errno) << endl);
            }
            DEBUG(2, "~vbs_state_type: " << monptr->first << " " << ((j==0)?("OK"):("FAIL")) << endl);
        }
        DEBUG(2, "~vbs_state_type: done" << endl);
    }
};

vbs_state_type vbs_state;


///////////////////////////////////////////////
//
//  Given a mountpoint, return a reference
//  to the mountpointhandler data
//
///////////////////////////////////////////////
mphandler_data& get_mpdata( string const& mp ) {
    mpmonitormap_type::iterator p = vbs_state.mpmonitorMap.find( mp );

    EZASSERT2( p!=vbs_state.mpmonitorMap.end(), vbs_except, EZINFO("Mountpoint " << mp << " not found in monitormap" << endl) );
    return p->second;
}


///////////////////////////////////////////////
//   Wrapper around ::pthread_create(3) -
//   creates a joinable thread with
//   ALL signals blocked.
//   Takes same arguments as ::pthread_create(3)
//   apart from the pthread_attr_t struct
///////////////////////////////////////////////
int vbs_pthread_create(pthread_t* thread, void *(*start_routine)(void*), void *arg) {
    int             pr, createrv;
    sigset_t        oldSig, newSig;
    pthread_attr_t  attr;

    // Make sure we have a joinable thread
    if( (pr=::pthread_attr_init(&attr))!=0 ) {
        DEBUG(-1, "vbs_pthread_create: pthread_attr_init fails - " << ::strerror(pr) << endl);
        return pr;
    }
    if( (pr=::pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE))!=0 ) {
        DEBUG(-1, "vbs_pthread_create: pthread_attr_setdetachstate fails - " << ::strerror(pr) << endl);
        return pr;
    }
    // Install a fully filled signal set (i.e. block all of 'm) and save the
    // current one
    if( sigfillset(&newSig)!=0 ) {
        DEBUG(-1, "vbs_pthread_create: sigfillset fails - " << ::strerror(errno) << endl);
        return errno;
    }
    if( (pr=::pthread_sigmask(SIG_SETMASK, &newSig, &oldSig))!=0 ) {
        DEBUG(-1, "vbs_pthread_create: pthread_sigmask (setting new mask) fails - " << ::strerror(pr) << endl);
        return pr;
    }
    // Now we're in a determined state and we can safely create the 
    // thread. We save the return value of this one specifically because
    // that will be the eventual return value. We do the cleanup calls
    // and if they'll fail we inform the user that but do not return *those*
    // error value(s)
    createrv = ::pthread_create(thread, &attr, start_routine, arg);
    if( createrv!=0 )
        DEBUG(-1, "vbs_pthread_create: pthread_create fails - " << ::strerror(createrv) << endl);

    // Cleanup phase: put back old signal mask & destroy pthread attributes
    if( (pr=::pthread_sigmask(SIG_SETMASK, &oldSig, 0))!=0 )
        DEBUG(-1, "vbs_pthread_create: pthread_sigmask (put back old mask) fails - " << ::strerror(pr) << endl);
    if( (pr=::pthread_attr_destroy(&attr))!=0 )
        DEBUG(-1, "vbs_pthread_create: pthread_attr_destroy fails - " << ::strerror(pr) << endl);

    // Phew. Finally done.
    return createrv;
}

///////////////////////////////////////////////
//  The mountpoint monitor function.
//  It is a thread function
///////////////////////////////////////////////

struct mapresult_type {
    //time_t           mtime;
    struct stat      st;
    scanfunction_ptr scanFunction;
    openpath_fn      openFunction;
    closepath_fn     closeFunction;

    mapresult_type():
        scanFunction( 0 ), openFunction( 0 ), closeFunction( 0 )
    { ::memset(&st, 0x0, sizeof(struct stat)); }

    mapresult_type(struct stat s, scanfunction_ptr sf, openpath_fn opener, closepath_fn closer):
        st( s ), scanFunction( sf ), openFunction( opener ), closeFunction( closer )
    {}

    operator bool() const {
        return !(scanFunction==0 || openFunction==0 || closeFunction==0);
    }
};

mapresult_type  mk6(struct stat t) {
    return mapresult_type(t, scanMk6RecordingFile, mk6_path_opener, mk6_path_closer);
}
mapresult_type  vbs(struct stat t) {
    return mapresult_type(t, scanVBSRecordingDirectory, vbs_path_opener, vbs_path_closer);
}


// predicate which checks if the given entry is a Mark6 data file
struct isMark6Recording {
    bool operator()(string const& entry) const {
        // attempt to read mk6 header
        int             fd = ::open(entry.c_str(), O_RDONLY);
        ssize_t         n;
        mk6_file_header fhdr;

        DEBUG(4, "isMark6Recording(" << entry << "): fd=" << fd << endl);
        if( fd<0 )
            return false;
        // attempt to read sizeof(header)
        n = ::read(fd, &fhdr, sizeof(mk6_file_header));
        ::close( fd );
        // Check if we could read the file header and wether it has the magicz
        return (n==sizeof(mk6_file_header) && fhdr.sync_word==MARK6_SG_SYNC_WORD && fhdr.version==2) ? true : false;
    }
};

// Complaint by users: vbs_ls, vbs_rm and vbs_fs don't seem to pick up
// FlexBuff recordings with regex majik characters (".", "+" et.al.) in
// them. Now, creating recordings with those characters in their names might
// be capital offence in the first place ... but since no-one's stopping
// them it might be considered polite to suck it up and make the s/w operate
// correctly just the same.
// In vbs_ls/vbs_rm the issue was fixed by escaping the recording name
// before making a regex pattern out of it for the VBS shrapnel:
//      https://docs.python.org/2/library/re.html#re.escape
//
// I've looked up the implementation of re.escape:
//      https://github.com/python/cpython/blob/master/Lib/re.py#L249
//      (status as on 09 Feb 2017)
//
// and I think it's easy enough to emulate that here in C++
static const string alphanum_str("_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890");

string escape(string const& s) {
    string                       rv;
    back_insert_iterator<string> output(rv);

    for(string::const_iterator p=s.begin(); p!=s.end(); *output++ = *p++)
        if( alphanum_str.find(*p)==string::npos )
            *output++ = '\\';
    return rv;
}

// predicate which tests if the given entry is a VBS style recording:
//    * entry represents a directory
//    * that directory contains at least one file <entry>.XXXXXXXX 
//      "entry/entry.XXXXXXXX is the pattern for a VBS recording chunk)

// predicate that returns wether or not a file name matches a recording chunk
struct isRecordingChunk {
    isRecordingChunk(string const& recname):
        __m_regex( string("(^|/)")+escape(recname)+"\\.[0-9]{8}$" )
    {}

    bool operator()(string const& entry) const {
        return __m_regex.matches(entry);
    }

    Regular_Expression  __m_regex;

    private:
        isRecordingChunk();
};

struct isVBSRecording {
    bool operator()(string const& entry) const {
        DEBUG(4, "isVBSRecording(" << entry << ")" << endl);
        // Test if any of the entries in this directory conform to the VBS
        // recording naming convention. As this is a quick scan we quit as
        // soon as we find one (1) matching entry. The full scan will be
        // done later
        // NOTE: 'entry' is the full path to a directory:
        //     /path/to/some/potential/vbs_recording_no0001
        // so we have to extract the last part of the path!
        try {
            return dir_exists(entry, isRecordingChunk(entry.substr(entry.rfind('/')+1)));
        }
        catch( std::exception& e ) {
            DEBUG(4, "isVBSRecording(" << entry << ")/caught exception: " << e.what() << endl);
            return false;
        }
        catch(...) {
            DEBUG(4, "isVBSRecording(" << entry << ")/caught unknown exception" << endl);
            return false;
        }
    }
};


// Define a directory mapper which we can map over a mountpoint.
// Each entry will be translated into a mapresult_type.
// Any entry with a non-empty mapresult will have to be added to the global
// recording cache ...
struct isRecording {
    // for each entry we return a value of this type
    typedef mapresult_type  value_type;

    // The actual implementation
    value_type operator()(string const& entry) const {
        // Ok it's a thing we recognize. Need to get the st_mtime
        // for the metadata
        struct stat  s;
        value_type   (*maker)(struct stat) = 0;

        if( ::lstat(entry.c_str(), &s)==0 ) {
            // Check if it's anything we recognize
            //  directories could be VBS recording,
            //  files could be Mark6 recording
            if( ((s.st_mode&S_IFDIR)==S_IFDIR) && isVBSRecording_f(entry) )
                maker = vbs;
            else if( ((s.st_mode&S_IFREG)==S_IFREG) && isMark6Recording_f(entry) )
                maker = mk6;
        } else {
            DEBUG(1, "WARN: Failed to stat(" << entry << ") - " << ::strerror(errno) << endl);
        }
        return (maker == 0) ? mapresult_type() : maker( s );
    }

    isMark6Recording    isMark6Recording_f;
    isVBSRecording      isVBSRecording_f;
};

struct name_matcher_type {
    name_matcher_type(string const& nm):
        __name_to_match( nm )
    {}

    bool operator()( const recording_type& rec ) const {
        return rec.recordingName==__name_to_match;
    }

    const string __name_to_match;
};


// note that we only ever ADD files to the chunk cache. If someone deletes
// the backing storage from out under us, that's too bad ...
void* mountpoint_monitor(void* args) {
    // For quick scanning the mount point
    typedef dir_mapper<isRecording>           recordingmapper_type;
    typedef recordingmapper_type::value_type  found_recordings_type;

    // We get passed a pointer to string
    bool            except = false;
    ostringstream   oss;
    mphandler_data* mpdata = (mphandler_data*)args;

    if( mpdata==0 ) {
        DEBUG(-1, "mountpoint_monitor/null-pointer passed as handler data!" << endl);
        return (void*)0;
    }

    if( mpdata->matchList.size() )
        oss << " [filtering " << mpdata->matchList.size() << " pattern" << ((mpdata->matchList.size()==1) ? "" : "s") << "]";
    DEBUG(2, "mountpoint_monitor[" << mpdata->mountPoint << "] starting" << oss.str() << endl);

    try {
        // Let's check that mother!
        // Do the scan unlocked
        found_recordings_type found = recordingmapper_type()(mpdata->mountPoint, *mpdata);
       
        // Filter out the entries that were diagnosed to be recordings.
        // We must hold the lock on the recordings.
        ::pthread_mutex_lock( &vbs_state.chunkMutex );
        recordings_type&    recordings( vbs_state.recordings );

        for(found_recordings_type::const_iterator ptr=found.begin(); ptr!=found.end(); ptr++) {
            mapresult_type const&  mr( ptr->second );

            // If it was not a recognized recording ..
            if( !mr )
                continue;

            // If recording does not exist yet, add it
            recordings_type::iterator recptr = std::find_if(recordings.begin(), recordings.end(), name_matcher_type(ptr->first));

            if( recptr==recordings.end() ) 
                recptr = recordings.insert(
                             recording_type(ptr->first, mr.scanFunction, mr.openFunction, mr.closeFunction)
                        ).first;
            // Ok, recording is certain to exist and recptr points to it
            // Now add path and update meta data
            recording_type const&   recording( *recptr );

            recording.add_path( mpdata->mountPoint+"/"+ptr->first );
            update(recording.metadata.st, ptr->second.st);
        }
        // Done updating the global recording map
        ::pthread_mutex_unlock( &vbs_state.chunkMutex );

        // Signal that we're done indexing
        ::pthread_mutex_lock(mpdata->mutex);
        mpdata->indexed = true;
        ::pthread_cond_signal(mpdata->condition);
        ::pthread_mutex_unlock(mpdata->mutex);

        // And drop into our inner loop!
        while( true ) {
            bool            stop    = false; // copies of shared state variables
            filechunk_type* workptr = 0;

            // lock self & condition wait for something to happen
            ::pthread_mutex_lock(mpdata->mutex);
            while( mpdata->stop==false && mpdata->workqueue.size()==0 )
                ::pthread_cond_wait(mpdata->condition, mpdata->mutex);

            stop = mpdata->stop;
            if( mpdata->workqueue.size() ) {
                workptr = mpdata->workqueue.front();
                mpdata->workqueue.pop();
            }
            ::pthread_mutex_unlock(mpdata->mutex);

            // Check what the message was
            if( stop==true )
                break;

            if( workptr==0 ) {
                DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] got nullptr in workqueue?!" << endl);
                continue;
            }
            // Crap. There's no denying anymore - we have to Do Work(tm)
            filechunk_type::read_chunk( workptr );
        }
        // Ok, we're not reading anymore. Inform all queued bloccks
        // that it ain't gonna happen no more
        ::pthread_mutex_lock(mpdata->mutex);
        while( mpdata->workqueue.size() ) {
            filechunk_type* fcptr = mpdata->workqueue.front();

            ::pthread_mutex_lock(fcptr->__m_mutex);
            fcptr->__m_thread_active = false;
            ::pthread_cond_broadcast(fcptr->__m_cond);
            ::pthread_mutex_unlock(fcptr->__m_mutex);
            mpdata->workqueue.pop();
        }
        ::pthread_mutex_unlock(mpdata->mutex);
    }
    catch( int eno ) {
        except = true;
        DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] caught errno " << eno << " - " << ::strerror(eno) << endl);
    }
    catch( std::exception& e ) {
        except = true;
        DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] caught exception " << e.what() << endl);
    }
    catch( ... ) {
        except = true;
        DEBUG(-1, "mountpoint_monitor[" << mpdata->mountPoint << "] caught unknown exception" << endl);
    }
    if( except ) {
        // Signal to main loop that (1) we're done initializing and (2)
        // that we don't accept any requests
        ::pthread_mutex_lock(mpdata->mutex);
        mpdata->indexed = true;
        mpdata->stop    = true;
        ::pthread_cond_signal(mpdata->condition);
        ::pthread_mutex_unlock(mpdata->mutex);
    }
    DEBUG(2, "mountpoint_monitor[" << mpdata->mountPoint << "] stopping" << endl);
    return (void*)0;
}


// Predicate for a directory entry to be a mountpoint
// 1. name must be "disk[0-9]+"
// 2. must refer to a directory
struct isMountpoint {
    bool operator()(string const& entry) const {
        Regular_Expression      rxDisk("^disk[0-9]{1,}$");
        struct stat             status;
        string::size_type       slash = entry.find_last_of("/");

        // IF there is a slash, we skip it, if there isn't, we
        // use the whole string
        if( slash==string::npos )
            slash = 0;
        else
            slash += 1;
        DEBUG(3, "isMountpoint: checking name " << entry.substr(slash) << endl);
        if( !rxDisk.matches(entry.substr(slash)) )
            return false;
        if( ::lstat(entry.c_str(), &status)<0 ) {
            DEBUG(2, "predMountpoint: ::lstat fails on " << entry << " - " << ::strerror(errno) << endl);
            return false;
        }
        // We must have r,x access to the directory [in order to descend into it]
        return S_ISDIR(status.st_mode) && (status.st_mode & S_IRUSR) && (status.st_mode & S_IXUSR);
    }
};


///////////////////////////////////////////////
//
//   Here follow the implementations 
//   of the file system functions
//   that we support
//
///////////////////////////////////////////////


void* vbs_init(struct fuse_conn_info* /*conn*/) {
    DEBUG(-1, "vbs_init: starting" << endl);
    try {
        mpmonitormap_type& mpmon( vbs_state.mpmonitorMap );

        EZASSERT2( vbs_state.mountPoints.size()>0, vbs_except, EZINFO("No mountpoints found") );

        for( paths_type::const_iterator mp=vbs_state.mountPoints.begin(); mp!=vbs_state.mountPoints.end(); mp++) {
            DEBUG(2, "vbs_init: Found mountpoint " << *mp << endl);

            // Excellent, now attempt to start a monitor thread for this mountpoint
            typedef pair<mpmonitormap_type::iterator, bool>  mpinsres_type;
            int            r;
            mpinsres_type  insres = mpmon.insert( make_pair(*mp, mphandler_data(*mp, vbs_state.indexPatterns)) );

            // Start the monitor thread
            if( (r=vbs_pthread_create(&insres.first->second.threadId, mountpoint_monitor, &insres.first->second))!=0 ) {
                DEBUG(-1, "vbs_init: Failed to create thread for mountpoint " << *mp << endl);
                // remove from mpmonitor map orelse we'll be waiting for a
                // thread that was never started
                // Again: actual killing of vbs_init() will happen later,
                // after we've waited for all other threads
                mpmon.erase( insres.first );
            }
        }

        // For each thread that was started, wait for it to have signalled it's done initializing
        DEBUG(-1, "vbs_init: please wait whilst indexing " << mpmon.size() << " mountpoints" << endl);
        for( mpmonitormap_type::iterator monptr=mpmon.begin(); monptr!=mpmon.end(); monptr++) {
            bool                 stop;
            mphandler_data&      mpdata = monptr->second;

            ::pthread_mutex_lock(mpdata.mutex);
            while( mpdata.stop==false && mpdata.indexed==false )
                ::pthread_cond_wait(mpdata.condition, mpdata.mutex);
            stop = mpdata.stop;
            ::pthread_mutex_unlock(mpdata.mutex);
            DEBUG(1, "vbs_init: " << monptr->first << " " << ((stop)?("EXIT"):("OK")) << endl);
        }

        DEBUG(-1, "vbs_init: indexing " << mpmon.size() << " mountpoints finished" << endl);


        // Only do the full index scan if requested
        recordings_type&    recordings( vbs_state.recordings );
        if( vbs_state.fullIndex ) {
            DEBUG(-1, "vbs_init: start indexing metadata of " << recordings.size() << " recordings ..." << endl);
            for(recordings_type::iterator recptr=recordings.begin(); recptr!=recordings.end(); recptr++)
                recptr->index();
        }
        DEBUG(-1, "vbs_init: indexing metadata finished" << endl);
    }
    catch( int erno ) {
        DEBUG(-1, "vbs_init/caught deadly errno - " << ::strerror(erno) << endl);
    }
    catch( const exception& e ) {
        DEBUG(-1, "vbs_init/caught deadly exception - " << e.what() << endl);
    }
    catch( ... ) {
        DEBUG(-1, "vbs_init/caught deadly unknown exception" << endl);
    }
    return (void*)&vbs_state;
}

int vbs_readdir(char const* path, void* buf, fuse_fill_dir_t filler, off_t /*offset*/, struct fuse_file_info* /*fi*/) {
    DEBUG(1, "vbs_readdir[" << path << "]" << endl);
    // We will *never* allow subdirectories so any request for anything
    // other than "/" (the root of this file system) we're going to return
    // an error!
    if( string(path)!="/" )
        return -ENOENT;
    // Now fill the directory listing
    filler(buf, ".",  0, 0);
    filler(buf, "..", 0, 0);

    // Grab the lock on the recordings and fetch all the keys - they are the
    // recording names!
    scoped_lock_type       sml( &vbs_state.chunkMutex );
    const recordings_type&  recordings = vbs_state.recordings;

    for(recordings_type::iterator recptr=recordings.begin(); recptr!=recordings.end(); recptr++)
        filler(buf, recptr->recordingName.c_str(), 0, 0);
    return 0;
}


int vbs_getattr(char const* path, struct stat* statbuf) {
    const time_t now = ::time(0);
    DEBUG(1, "vbs_getattr[" << path << "]" << endl);
    // before we create the recording name out of the path,
    // we have to strip leading slashes!
    while( *path=='/' ) path++;
    const string recname( path );
  
    // If someone requests the attrs of "/" it's the props of
    // the FUSE mountpoint itself, so we list it as a directory 
    // [note, after stripping leading slashes, the root path
    //  becomes the empty string!]
    if( recname=="" ) {
        statbuf->st_mode  = S_IFDIR | permissionBits;
        statbuf->st_uid   = ::getuid();
        statbuf->st_gid   = ::getgid();
        statbuf->st_mtime = now;
        statbuf->st_ctime = now;
        statbuf->st_atime = now;
        return 0;
    }
    scoped_lock_type                sml( &vbs_state.chunkMutex );
    recordings_type&                recordings( vbs_state.recordings );
    recordings_type::const_iterator recptr = std::find_if(recordings.begin(), recordings.end(), name_matcher_type(recname));

    if( recptr==vbs_state.recordings.end() ) {
        DEBUG(-1, "vbs_getattr[" << path << "] - not found in cache?!!!" << endl);
        return -ENOENT;
    }
    // It's an existing recording!
    // should get mode/user/group from the actual file(s)-on-disk
    struct stat const&   recStat(recptr->metadata.st);
    statbuf->st_mode  = recStat.st_mode;
    statbuf->st_uid   = recStat.st_uid;
    statbuf->st_gid   = recStat.st_gid;
    statbuf->st_size  = recptr->metadata.recordingSize;
    statbuf->st_mtime = statbuf->st_ctime = statbuf->st_atime = recStat.st_mtime;

    DEBUG(3, "vbs_getattr[" << path << "]: " << *statbuf << endl);
    return 0;
}

void vbs_destroy(void* /*ptr*/) { }


int vbs_open(char const* path, struct fuse_file_info* fi) {
    // Skip leading slash(es) in path
    while( path && *path && *path=='/' )
        path++;

    int                       res = 0;
    const string              recname( path );

    ::pthread_mutex_lock(&vbs_state.chunkMutex);

    recordings_type&          recordings = vbs_state.recordings;
    openfilecache_type&       openfiles  = vbs_state.openfileCache;
    recordings_type::iterator recptr     = std::find_if(recordings.begin(), recordings.end(), name_matcher_type(recname));

    if( recptr==recordings.end() ) {
        res = -ENOENT;
    } else {
        // Find first unused file descriptor
        struct rlimit   rlim;

        res    = -EMFILE; // will be reset to succes if indeed we find a free fd
        fi->fh = std::numeric_limits<int>::max();

        if( ::getrlimit(RLIMIT_NOFILE, &rlim)!=0 ) {
            DEBUG(-1, "::getrlimit(RLIMIT_NOFILE, ...) fails - " << ::strerror(errno) << endl);
            rlim.rlim_max = rlim.rlim_cur = std::numeric_limits<int>::max() - 1024;
        }

        while( (fi->fh>rlim.rlim_max) && openfiles.find(fi->fh)!=openfiles.end() )
            fi->fh--;
        if( fi->fh>=rlim.rlim_max ) {
            // do not forget to open the recording
            recptr->open( (int)fi->fh );
            openfiles.insert( make_pair(fi->fh, openfile_type(recptr)) );
            res  = 0; // reset to success
        } 
    }
    ::pthread_mutex_unlock(&vbs_state.chunkMutex);
    DEBUG(1, "vbs_open[" << path << "] res=" << res << " fh=" << fi->fh << endl);

    return res;
}


int vbs_read(char const* path, char* buff, size_t n, off_t offset, struct fuse_file_info* fi) {
    static unsigned int nRead = 0;
    unsigned int        nr = nRead++;
    openfilecache_type::iterator ofp;
   
    {
        // make use of the fact that once we have an iterator
        // into a std::map or std::set, other inserts/deletes won't
        // invalidate this iterator. Only someone else doing a removal
        // of this entry will invalidate the iterator we got (if any)
       scoped_lock_type sml( &vbs_state.chunkMutex );
       
       if( (ofp=vbs_state.openfileCache.find(fi->fh))==vbs_state.openfileCache.end() )
           return -ENOENT;
    }
    // Lock the open file - we must have exclusive access to of->curChunk;
    // either we need to modify it andalso we must prevent others from
    // modifying it behind our backs
    openfile_type&            of = ofp->second;
    scoped_writelock_type     swl( &ofp->second.fileRWLock );
    int                       eno  = 0, racnt;
    unsigned int              subn = 0, priority;
    size_t                    navail;
    const size_t              norg = n;
    filechunks_type&          chunks = of.recordingPtr->fileChunks;
    filechunks_type::iterator ptr = of.curChunk, raptr;

    // Count consecutive reads and set priority depending on that
    if( offset==of.previousEnd )
        of.nConsecutiveReads++;
    else
        of.nConsecutiveReads = 0;
    // Threshold may be subject to change - maybe fuse does some readahead
    // by itself, even if the user only asks for a tiny bit once
    priority = (of.nConsecutiveReads>3);

    // If data is not from the current block [someone changed the file
    // pointer behind our backs through e.g. lseek(2)] we must unbobserve
    // blocks we don't use no more [the ones for which we issued
    // read-aheads]
    if( ptr==chunks.end() || (ptr!=chunks.end() && ptr->remainingBytes(offset)==0) ) {
        filechunks_type::iterator rmptr = chunks.begin();
        // release anything up to the block we *need* to have
        while( rmptr!=chunks.end() && rmptr->remainingBytes(offset)==0 ) {
            rmptr->releaseData(ofp->first);
            rmptr++;
        }
        // Save pointer to chunk we need to be a-readin from
        of.curChunk = ptr = rmptr;
        // Attempt to advance "readahead" nr of blocks - those
        // we'll potentially need so we leave their in-memory state
        // as-is. If prio==0 this means no read-ahead actually.
        for(racnt=0; priority>0 && racnt<=vbs_state.readAhead && rmptr!=chunks.end(); racnt++, rmptr++);
        // Release anything up to end of file, even though we possibly need
        // those again later on but such is life in cache-land; our
        // recordings are so big that we cannot possibly keep everything in
        // memory that we want. Also: it is the application's responsibility
        // to not seek to and fro throughout the recording!
        while( rmptr!=chunks.end() ) {
            rmptr->releaseData(ofp->first);
            rmptr++;
        }
    }
    while( n ) {
        DEBUG(3, "vbs_read#" << nr << "." << subn++ << "[" << path << ", " << ::pthread_self() << "] " << 
                 "offset=" << offset << " n=" << n << " prio=" << priority << " prevEnd=" << of.previousEnd << " nConsec=" << of.nConsecutiveReads << endl);
        // if no more bytes available, move to next chunk
        for( ;ptr!=chunks.end() && (navail=ptr->remainingBytes(offset))==0; ptr++)
            ptr->releaseData(ofp->first);

        // either end of file or point at chunk that we need to read from
        // attempt to read past end of file is not an error
        if( ptr==chunks.end() )
            break;

        // We now point at block that we need data from so issue read for
        // this block as well as the following nReadAhead - depending on
        // priority. If 
        for(raptr=ptr, racnt=0; racnt<=(priority>0 ? vbs_state.readAhead : 0) && raptr!=chunks.end(); racnt++, raptr++)
            raptr->initiateRead( ofp->first, priority );

        // We know that navail>0 at this point
        const size_t    n2r = std::min(n, navail);
        unsigned char*  data = ptr->readLockForData();

        if( data==0 ) {
            eno = errno;
            DEBUG(-1, "vbs_read: " << ptr->pathToChunk << " FAILS: " << ::strerror(errno) << endl);
            break;
        }

        // Ok data available!
        ::memcpy(buff, data+(offset - ptr->chunkOffset), n2r);

        buff   += n2r;
        n      -= n2r;
        offset += n2r;
        // Indicate that we're done with this chunk for now
        ptr->releaseReadLock();
    }
    of.curChunk    = ptr;
    of.previousEnd = offset;
    DEBUG(3, "vbs_read#" << nr << "." << subn << "[" << path << ", " << ::pthread_self() << "] "
             "offset=" << offset << " n=" << n << " errno=" << eno << " FINISHED; rv=" << (eno ? -eno : (norg - n)) << endl);

    return eno ? -eno : (norg - n);
}

int vbs_release(char const* path, struct fuse_file_info* fi) {
    scoped_lock_type              slt( &vbs_state.chunkMutex );
    openfilecache_type::iterator  ofp = vbs_state.openfileCache.find( fi->fh );

    DEBUG(1, "vbs_release[" << path << "] fh=" << fi->fh << endl);

    if( ofp==vbs_state.openfileCache.end() )
        return -ENOENT;

    filechunks_type&     chunks = ofp->second.recordingPtr->fileChunks;

    // Wait for issued chunk-read-requests to finish and properly clean them up
    for(filechunks_type::iterator fcptr = chunks.begin(); fcptr!=chunks.end(); fcptr++) {
        if( fcptr->readLockForData() )
            fcptr->releaseReadLock();
        fcptr->releaseData( ofp->first );
    }
    // Do not forget to close the recording for this observer
    ofp->second.recordingPtr->close( ofp->first );
    DEBUG(1, "vbs_release[" << path << "] done" << endl);
    vbs_state.openfileCache.erase( ofp );
    return 0;
}

///////////////////////////////////////////////
//   A filesystem must define a set
//   of operations with callback functions
//   on the file system objects it
//   represents. All callbacks are 
//   registered in the fuse_operations struct
///////////////////////////////////////////////
struct vbs_fuse_operations :
    public fuse_operations {

        vbs_fuse_operations() {
            init    = &vbs_init;
            destroy = &vbs_destroy;
            getdir  = NULL;
            readdir = &vbs_readdir;
            getattr = &vbs_getattr;
            open    = &vbs_open;
            read    = &vbs_read;
            release = &vbs_release; // called on close(2)
        }
};

vbs_fuse_operations vbs_oper;
#if 0
struct fuse_operations vbs_oper = {
    //.getattr = vbs_getattr,
    //.readdir = vbs_readdir,
    //.open = vbs_open,
    //.read = vbs_read,
    //.readlink = vbs_readlink,
    // no .getdir -- that's deprecated
    //.getdir = NULL,
    //.mknod = vbs_mknod,
    //.mkdir = vbs_mkdir,
    //.unlink = vbs_unlink,
    //.rmdir = vbs_rmdir,
    //.symlink = vbs_symlink,
    //.rename = vbs_rename,
    //.link = vbs_link,
    //.chmod = vbs_chmod,
    //.chown = vbs_chown,
    //.truncate = vbs_truncate,
    //.utime = vbs_utime,
    //.write = vbs_write,
    /** Just a placeholder, don't set */ // huh???
    //.statfs = vbs_statfs,
    //.flush = vbs_flush,
    //.release = vbs_release,
    //.fsync = vbs_fsync,
    //.setxattr = vbs_setxattr,
    //.getxattr = vbs_getxattr,
    //.listxattr = vbs_listxattr,
    //.removexattr = vbs_removexattr,
    //.opendir = vbs_opendir,
    //.releasedir = vbs_releasedir,
    //.fsyncdir = vbs_fsyncdir,
    .init = vbs_init,
    .destroy = vbs_destroy,
    //.access = vbs_access,
    //.create = vbs_create,
    //.ftruncate = vbs_ftruncate,
};
#endif


/////////////////////////////////////////////////////////////////
//
//  Transform a list of globbety glob shell patterns
//  into a set of directories such that we don't have
//  duplicates
//
////////////////////////////////////////////////////////////////
paths_type globbety_glob(patternlist_type const& patterns) {
    int         flags = GLOB_NOSORT;
    paths_type  retval;

#ifdef GLOB_BRACE
    flags |= GLOB_BRACE;
#endif

    for(patternlist_type::const_iterator p=patterns.begin(); p!=patterns.end(); p++) {
        // globbety glob
        glob_t      globs;
        struct stat st;

        if( ::glob(p->c_str(), flags, 0, &globs)==0 ) {
            // ok, we found matches, only add the ones that are directories
            for(size_t i=0; i<globs.gl_pathc; i++)
                if( ::lstat(globs.gl_pathv[i], &st)==0 && (st.st_mode&S_IFDIR)==S_IFDIR )
                    retval.insert( globs.gl_pathv[i] );
        } else {
            DEBUG(-1, "glob(" << *p << ") fails - " << ::strerror(errno) << endl);
        }
        ::globfree(&globs);
    }
    return retval;
}



/////////////////////////////////////////////////////////////////
//   Command line handling
//   There's only a few options
//   that we support
//    -h            print help and exit
//    --quick       do not do a full index
//    --version     print version and exit
//    -m <int>      setting debug level [higher = more output]
//    -I <pattern>  add pattern to list of recordings to index
//    -R <pattern>  add pattern to list of mount points to index
//    -v            add flexbuff pattern to list of paths
//    -6            add Mark6 pattern to list of paths
//    -f            FUSE 'foreground' flag
//
////////////////////////////////////////////////////////////////
struct vbs_settings {
    int              read_ahead;
    int              debug_level;
    bool             help;
    bool             version;
    bool             fullScan;
    patternlist_type indexPatterns;
    patternlist_type patterns;

    // Initialize with defaults
    vbs_settings():
        read_ahead( 0 ), debug_level( 0 ), help( false ), version( false ), fullScan( true )
    {}
};

ostream& operator<<(ostream& os, const vbs_settings& vbo) {
    os << "-m " << vbo.debug_level << " ";
    if( vbo.version )
        os << "--version ";
    if( !vbo.fullScan )
        os << "--quick ";
    for(patternlist_type::const_iterator pattern=vbo.patterns.begin(); pattern!=vbo.patterns.end(); pattern++)
        os << "-R " << *pattern << " ";
    for(patternlist_type::const_iterator pattern=vbo.indexPatterns.begin(); pattern!=vbo.indexPatterns.end(); pattern++)
        os << "-I " << *pattern << " ";
    if( vbo.read_ahead>0 )
        os << "-n " << vbo.read_ahead << " ";
    if( vbo.help )
        os << "-h ";
    return os;
}

#define OFFSETOF(tp, fld) \
    (size_t)((unsigned char*)(&(((tp*)1)->fld)) - (unsigned char*)1)
#define VBSFS_OPT(t, p, v) { t, OFFSETOF(struct vbs_settings, p), v }

struct fuse_opt vbs_options[] = {
    VBSFS_OPT("-m %i", debug_level, 0),
    VBSFS_OPT("-n %i", read_ahead,  0),
    // For these we require specific processing [neither are just int or
    // char* options]
    {"-I %s"    , -1U, 0x4A59},  // 'JY' - see help on the options ;-)
    {"-R %s"    , -1U, 42},      //
    {"-6"       , -1U, 666},     // 
    {"-v"       , -1U, 112},     //  All of these numbers 
    {"--help"   , -1U, 911},     //  were randomly chosen, obviously!
    {"--version", -1U, 31415926},//
    {"--quick"  , -1U, 27182818},//
    // Sentinel
    {NULL, 0, 0}
};

int vbs_option_proc(void* data, const char* arg, int key, struct fuse_args* /*outargs*/) {
    int           rv = 0; // discard by default
    const string  arg_s( arg );
    vbs_settings* settingsptr = (vbs_settings*)data;

    DEBUG(2, "vbs_option_proc[key=" << key << " (non_opt=" << FUSE_OPT_KEY_NONOPT << ")]: processing " << arg_s << endl);
    switch( key ) {
        case 31415926:
            settingsptr->version  = true;
            break;
        case 27182818:
            settingsptr->fullScan = false;
            break;
        case 0x4A59:
            // we know that the arg looks like "-I<pattern>"
            settingsptr->indexPatterns.push_back( arg_s.substr(2) );
            break;
        case 911:
            settingsptr->help     = true;
            break;
        case 112:
            settingsptr->patterns.push_back( flexbuffPattern );
            break;
        case 666:
            settingsptr->patterns.push_back( mk6Pattern );
            break;
        case 42:
            // we know that the arg looks like "-R<pattern>"
            settingsptr->patterns.push_back( arg_s.substr(2) );
            break;
        default:
            // we didn't handle the argument so let's keep it!
            rv = 1;
            break;
    }
    return rv;
}

void Usage(string const& prog ) {
    cout << "Usage: " << prog << " [options] /path/to/mountpoint" << endl << endl;
    cout << "VBS specific command line arguments:" << endl << endl
         << "   -f           run in foreground/do not background" << endl
         << "   --help       print this message and exit" << endl
         << "   --version    print version and exit" << endl
         << "   --quick      do not do a full index at startup" << endl
         << "                this will perform a quick scan of recordings" << endl
         << "                but full meta data (modification time, size, etc)" << endl
         << "                for a recording will only be gathered (and cached)" << endl
         << "                if is opened" << endl
         << "   -m <int>     set VBS_FS debug print level to <n>" << endl
         << "                higher <n> => more output, default 0 (no output)" << endl
         << "   -n <int>     issue <int> chunks read-ahead. Only useful if files" << endl
         << "                are read in strict linear fashion, otherwise it will" << endl
         << "                sigificantly HURT performance (default '1')" << endl
         << "   -I <pattern> only index recordings matching <pattern>" << endl
         << "                This option can be present multiple times." << endl
         << "                Shell-style wildcard characters are supported;" << endl
         << "                the implementation uses fnmatch(3) for matching." << endl
         << "                [credits to Jun Yang(Onsala) for requesting this]" << endl
         << "   -R <pattern> add pattern to list of directories to index" << endl
         << "                if nothing specified, defaults to /mnt/disk*" << endl
         << "                Note: the -R option can be specified multiple times" << endl
#ifdef GLOB_BRACE
         << "                Note: \"{pat1, pat2}\" syntax is supported in the pattern" << endl
#endif
         << "   -6           shorthand for adding the Mark6 directories to the" << endl
         << "                search path: /mtn/disks/*/*/data" << endl
         << "   -v           explicitly add the FlexBuff disk pattern to the list" << endl
         << "                of paths to search" << endl;
}



///////////////////////////////////////////////
//
//         The main function
//
///////////////////////////////////////////////
int main(int argc, char** argv) {
    // Before doing _anything_ check if we're not root/suid root
    if( ::getuid()==0 || ::geteuid()==0 ) {
        DEBUG(-1, "Running vbs_fuse as root opens unnacceptable security holes" << endl);
        return -11;
    }

    // Now we can safely start parsing
    int                 fuse_stat;
    vbs_settings        settings;
    struct fuse_args    args = FUSE_ARGS_INIT(argc, argv);

    // Do command line parsing
    ::fuse_opt_parse(&args, (void*)&settings, &vbs_options[0], vbs_option_proc);
    dbglev_fn( settings.debug_level );
    DEBUG(1, "fuse_opt_arg done [" << settings << "]" << endl);

    if( settings.version ) {
        cout << "$Id: vbs_fs.cc 53 2017-03-19 12:56:19Z jive_cc $" << endl;
        exit( 0 );
    }

    if( settings.help ) {
        string            prog( argv[0] );
        string::size_type slash = prog.rfind('/');

        Usage( (slash==string::npos) ? prog : prog.substr(slash+1) );
        exit( 0 );
    }

    // Set read-ahead value - only if it's >0
    if( settings.read_ahead<0 ) {
        DEBUG(-1, "invalid value " << settings.read_ahead << " for read ahead!" << endl);
        return -EINVAL;
    }

    // If no root pattern given, default to flexbuff pattern
    if( settings.patterns.empty() ) {
        DEBUG(-1, "WARN: No root dir specified for disks, defaulting to " << flexbuffPattern << endl);
        settings.patterns.push_back( flexbuffPattern );
    }
    // use glob.glob to expand all of the patterns given on the cmd line
    vbs_state.mountPoints = globbety_glob( settings.patterns );

    if( settings.read_ahead!=vbs_state.readAhead )
        vbs_state.readAhead = settings.read_ahead;

    // Do not forget to copy the --quick option 8-/
    vbs_state.fullIndex = settings.fullScan;

    // And transform the list-of-patterns into set-of-patterns to avoid duplicates :-)
    std::copy(settings.indexPatterns.begin(), settings.indexPatterns.end(),
              std::insert_iterator<paths_type>(vbs_state.indexPatterns, vbs_state.indexPatterns.end()));

    // Now it's about time to drop into FUSE's main loop
    DEBUG(3, "dropping into fuse_main" << endl);
    fuse_stat = fuse_main(args.argc, args.argv, &vbs_oper, &vbs_state);

    DEBUG(3, "fuse_stat returned " << fuse_stat << endl);
    return fuse_stat;
}


void scanVBSRecordingDirectory(string const& recname, string const& dir, filechunks_type& rv, struct stat& ts) {
    DIR*             dirp;
    struct stat      st;
    const string     mp(dir.substr(0, dir.rfind('/')));
    direntries_type  chunks;
    isRecordingChunk predicate( recname );

    if( (dirp=::opendir(dir.c_str()))==0 ) {
        DEBUG(4, "scanVBSRecordingDirectory(" << recname << ")/ ::opendir(" << dir << ")  fails - " << ::strerror(errno) << endl);
        return;
    }
    chunks = dir_filter(dirp, predicate);
    ::closedir(dirp);

    // If we find duplicates, now *that* is a reason to throw up
    for(direntries_type::const_iterator p=chunks.begin(); p!=chunks.end(); p++) {
        const string  fnam( dir + "/" + *p );
        if( ::lstat(fnam.c_str(), &st)!=0 ) {
            DEBUG(4, "scanVBSRecordingDirectory(" << recname << ")/ ::lstat(" << fnam << ") fails - " << ::strerror(errno) << endl);
            continue;
        }
        update(ts, st);

        // parse the chunk sequence number. Note that "*p" matched the
        // isRecordingChunk predicate, i.e. looks like "<recname>.XXXXXXXX"
        string::size_type   dot = p->rfind('.');
        const unsigned long seqno = ::strtoul(p->substr(dot+1).c_str(), 0, 10);

        EZASSERT2((rv.insert(filechunk_type(mp, fnam, (unsigned int)seqno, st.st_size))).second,
                  vbs_except, EZINFO(" duplicate insert for chunk " << *p));
    }
}

void scanMk6RecordingFile(string const& /*recname*/, string const& file, filechunks_type& rv, struct stat& ts) {
    int               fd;
    off_t             fpos;
    struct stat       st;
    const string      mp(file.substr(0, file.rfind('/')));
    const size_t      fh_size = sizeof(mk6_file_header);
    const size_t      wb_size = sizeof(mk6_wb_header_v2);
    unsigned char     buf[ fh_size+wb_size  ];
    mk6_file_header*  fh6  = (mk6_file_header*)&buf[0];
    mk6_wb_header_v2* wbh  = (mk6_wb_header_v2*)&buf[0];

    // File existence has been checked before so now we MUST be able to open it
    ASSERT2_POS( fd=::open(file.c_str(), O_RDONLY), SCINFO(" failed to open file " << file) );

    // It may well not be a Mk6 recording, for all we know
    if( ::read(fd, fh6, sizeof(mk6_file_header))!=sizeof(mk6_file_header) ) {
        DEBUG(4, "scanMk6RecordingFile[" << file << "]: fail to read mk6 header - " << ::strerror(errno) << endl);
        ::close(fd);
        return;
    }

    if( ::fstat(fd, &st)!=0 ) {
        DEBUG(4, "scanMk6RecordingFile[" << file << "]: fail to ::fstat - " << ::strerror(errno) << endl);
    } else {
        update(ts, st);
    }

    DEBUG(4, "scanMk6RecordingFile[" << file << "]: starting" << endl);
    // Ok. Now we should just read all the blocks in this file!
    fpos = fh_size;
    while( ::read(fd, wbh, wb_size)==(ssize_t)wb_size ) {
        // Don't forget that the block sizes written in the Mark6 files are
        // including the write-block-header size! (Guess how I found out
        // that I'd forgotten just that ...)

        // Make sure there's sense in the block number and size, otherwise better give up
        EZASSERT2(wbh->blocknum>=0 && wbh->wb_size>0, vbs_except,
                  EZINFO(" found bogus stuff in write block header @" << fpos << " in " << file <<
                         ", block# " << wbh->blocknum << ", sz=" << wbh->wb_size);
                  ::close(fd));

        // Ok, found another block!
        fpos += wb_size;

        // We cannot tolerate duplicate inserts
        EZASSERT2(rv.insert(filechunk_type(mp, file, (unsigned int)wbh->blocknum, wbh->wb_size-wb_size, fpos)).second, vbs_except,
                  EZINFO(" duplicate insert for chunk " << wbh->blocknum); ::close(fd) );

        // Advance file pointer
        fpos += (wbh->wb_size - wb_size);
        if( ::lseek(fd, fpos, SEEK_SET)==(off_t)-1 ) {
            DEBUG(4, "scanMk6RecordingFile[" << file << "]: failed to seek to next block @" <<
                     fpos << " - " << ::strerror(errno) << endl);
            break;
        }
    }
    // Actually, why don't I just close the file when done?
    // Thanks for AlexanderN and ChristianP for reporting the issue ...
    ::close(fd);
    DEBUG(4, "scanMk6RecordingFile[" << file << "]: done" << endl);
}
